/* -*-c-*-
 *
 * $Id: mpmont.c,v 1.1 1999/11/17 18:02:16 mdw Exp $
 *
 * Montgomery reduction
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: mpmont.c,v $
 * Revision 1.1  1999/11/17 18:02:16  mdw
 * New multiprecision integer arithmetic suite.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "mp.h"
#include "mpmont.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @mpmont_create@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *m@ = modulus to use
 *
 * Returns:	---
 *
 * Use:		Initializes a Montgomery reduction context ready for use.
 */

void mpmont_create(mpmont *mm, mp *m)
{
  /* --- Take a copy of the modulus --- */

  mp_shrink(m);
  mm->m = MP_COPY(m);

  /* --- Find the magic value @mi@ --- *
   *
   * This is a slightly grungy way of solving the problem, but it does work.
   */

  {
    mpw av[2] = { 0, 1 };
    mp a, b;
    mp *i;
    mpw mi;

    mp_build(&a, av, av + 2);
    mp_build(&b, m->v, m->v + 1);
    mp_gcd(0, 0, &i, &a, &b);
    mi = i->v[0];
    if (!(i->f & MP_NEG))
      mi = MPW(-mi);
    mm->mi = mi;
    MP_DROP(i);
  }

  /* --- Discover the values %$R \bmod m$% and %$R^2 \bmod m$% --- */

  {
    size_t l = MP_LEN(m);
    mp *r = mp_create(l + 1);

    mm->shift = l * MPW_BITS;
    MPX_ZERO(r->v, r->vl - 1);
    r->vl[-1] = 1;
    mm->r = mm->r2 = MP_NEW;
    mp_div(0, &mm->r, r, m);
    r = mp_sqr(r, mm->r);
    mp_div(0, &mm->r2, r, m);
    MP_DROP(r);
  }
}

/* --- @mpmont_destroy@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to a Montgomery reduction context
 *
 * Returns:	---
 *
 * Use:		Disposes of a context when it's no longer of any use to
 *		anyone.
 */

void mpmont_destroy(mpmont *mm)
{
  MP_DROP(mm->m);
  MP_DROP(mm->r);
  MP_DROP(mm->r2);
}

/* --- @mpmont_reduce@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = destination
 *		@const mp *a@ = source, assumed positive
 *
 * Returns:	Result, %$a R^{-1} \bmod m$%.
 */

mp *mpmont_reduce(mpmont *mm, mp *d, const mp *a)
{
  mpw *dv, *dvl;
  const mpw *mv, *mvl;
  size_t n;

  /* --- Initial conditioning of the arguments --- */

  n = MP_LEN(mm->m);

  if (d == a)
    MP_MODIFY(d, 2 * n);
  else {
    MP_MODIFY(d, 2 * n);
    memcpy(d->v, a->v, MPWS(MP_LEN(a)));
    memset(d->v + MP_LEN(a), 0, MPWS(MP_LEN(d) - MP_LEN(a)));
  }
    
  dv = d->v; dvl = d->vl;
  mv = mm->m->v; mvl = mm->m->vl;

  /* --- Let's go to work --- */

  while (n--) {
    mpw u = MPW(*dv * mm->mi);
    MPX_UMLAN(dv, dvl, mv, mvl, u);
    dv++;
  }

  /* --- Done --- */

  memmove(d->v, dv, MPWS(dvl - dv));
  d->vl -= dv - d->v;
  MP_SHRINK(d);
  d->f = a->f & MP_BURN;
  return (d);
}

/* --- @mpmont_mul@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = destination
 *		@const mp *a, *b@ = sources, assumed positive
 *
 * Returns:	Result, %$a b R^{-1} \bmod m$%.
 */

mp *mpmont_mul(mpmont *mm, mp *d, const mp *a, const mp *b)
{
  mpw *dv, *dvl;
  const mpw *av, *avl;
  const mpw *bv, *bvl;
  const mpw *mv, *mvl;
  mpw y;
  size_t n, i;

  /* --- Initial conditioning of the arguments --- */

  if (MP_LEN(a) > MP_LEN(b)) {
    const mp *t = a; a = b; b = t;
  }
  n = MP_LEN(mm->m);
    
  MP_MODIFY(d, 2 * n + 1);
  dv = d->v; dvl = d->vl;
  MPX_ZERO(dv, dvl);
  av = a->v; avl = a->vl;
  bv = b->v; bvl = b->vl;
  mv = mm->m->v; mvl = mm->m->vl;
  y = *bv;

  /* --- Montgomery multiplication phase --- */

  i = 0;
  while (i < n && av < avl) {
    mpw x = *av++;
    mpw u = MPW((*dv + x * y) * mm->mi);
    MPX_UMLAN(dv, dvl, bv, bvl, x);
    MPX_UMLAN(dv, dvl, mv, mvl, u);
    dv++;
    i++;
  }

  /* --- Simpler Montgomery reduction phase --- */

  while (i < n) {
    mpw u = MPW(*dv * mm->mi);
    MPX_UMLAN(dv, dvl, mv, mvl, u);
    dv++;
    i++;
  }

  /* --- Done --- */

  memmove(d->v, dv, MPWS(dvl - dv));
  d->vl -= dv - d->v;
  MP_SHRINK(d);
  d->f = (a->f | b->f) & MP_BURN;
  return (d);
}

/* --- @mpmont_exp@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@const mp *a@ = base
 *		@const mp *e@ = exponent
 *
 * Returns:	Result, %$a^e \bmod m$%.
 */

mp *mpmont_exp(mpmont *mm, const mp *a, const mp *e)
{
  mpscan sc;
  mp *ar = mpmont_mul(mm, MP_NEW, a, mm->r2);
  mp *d = MP_COPY(mm->r);

  mp_scan(&sc, e);

  if (MP_STEP(&sc)) {
    for (;;) {
      mp *dd;
      if (MP_BIT(&sc)) {
	dd = mpmont_mul(mm, MP_NEW, d, ar);
	MP_DROP(d);
	d = dd;
      }
      if (!MP_STEP(&sc))
	break;
      dd = mpmont_mul(mm, MP_NEW, ar, ar);
      MP_DROP(ar);
      ar = dd;
    }
  }
  MP_DROP(ar);
  return (mpmont_reduce(mm, d, d));
}

/*----- Test rig ----------------------------------------------------------*/

#ifdef TEST_RIG

static int tcreate(dstr *v)
{
  mp *m = *(mp **)v[0].buf;
  mp *mi = *(mp **)v[1].buf;
  mp *r = *(mp **)v[2].buf;
  mp *r2 = *(mp **)v[3].buf;

  mpmont mm;
  int ok = 1;

  mpmont_create(&mm, m);

  if (mm.mi != mi->v[0]) {
    fprintf(stderr, "\n*** bad mi: found %lu, expected %lu",
	    (unsigned long)mm.mi, (unsigned long)mi->v[0]);
    fputs("\nm = ", stderr); mp_writefile(m, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  if (MP_CMP(mm.r, !=, r)) {
    fputs("\n*** bad r", stderr);
    fputs("\nm = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\nexpected ", stderr); mp_writefile(r, stderr, 10);
    fputs("\n   found ", stderr); mp_writefile(r, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  if (MP_CMP(mm.r2, !=, r2)) {
    fputs("\n*** bad r2", stderr);
    fputs("\nm = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\nexpected ", stderr); mp_writefile(r2, stderr, 10);
    fputs("\n   found ", stderr); mp_writefile(r2, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  MP_DROP(m);
  MP_DROP(mi);
  MP_DROP(r);
  MP_DROP(r2);
  mpmont_destroy(&mm);
  return (ok);
}

static int tmul(dstr *v)
{
  mp *m = *(mp **)v[0].buf;
  mp *a = *(mp **)v[1].buf;
  mp *b = *(mp **)v[2].buf;
  mp *r = *(mp **)v[3].buf;
  mp *mr, *qr;
  int ok = 1;

  mpmont mm;
  mpmont_create(&mm, m);

  {
    mp *ar = mpmont_mul(&mm, MP_NEW, a, mm.r2);
    mp *br = mpmont_mul(&mm, MP_NEW, b, mm.r2);
    mr = mpmont_mul(&mm, MP_NEW, ar, br);
    mr = mpmont_reduce(&mm, mr, mr);
    MP_DROP(ar); MP_DROP(br);
  }

  qr = mp_mul(MP_NEW, a, b);
  mp_div(0, &qr, qr, m);

  if (MP_CMP(qr, !=, r)) {
    fputs("\n*** classical modmul failed", stderr);
    fputs("\n m = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\n a = ", stderr); mp_writefile(a, stderr, 10);
    fputs("\n b = ", stderr); mp_writefile(b, stderr, 10);
    fputs("\n r = ", stderr); mp_writefile(r, stderr, 10);
    fputs("\nqr = ", stderr); mp_writefile(qr, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  if (MP_CMP(mr, !=, r)) {
    fputs("\n*** montgomery modmul failed", stderr);
    fputs("\n m = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\n a = ", stderr); mp_writefile(a, stderr, 10);
    fputs("\n b = ", stderr); mp_writefile(b, stderr, 10);
    fputs("\n r = ", stderr); mp_writefile(r, stderr, 10);
    fputs("\nmr = ", stderr); mp_writefile(mr, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  MP_DROP(m);
  MP_DROP(a);
  MP_DROP(b);
  MP_DROP(r);
  MP_DROP(mr);
  MP_DROP(qr);
  mpmont_destroy(&mm);
  return ok;
}

static int texp(dstr *v)
{
  mp *m = *(mp **)v[0].buf;
  mp *a = *(mp **)v[1].buf;
  mp *b = *(mp **)v[2].buf;
  mp *r = *(mp **)v[3].buf;
  mp *mr;
  int ok = 1;

  mpmont mm;
  mpmont_create(&mm, m);

  mr = mpmont_exp(&mm, a, b);

  if (MP_CMP(mr, !=, r)) {
    fputs("\n*** montgomery modexp failed", stderr);
    fputs("\n m = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\n a = ", stderr); mp_writefile(a, stderr, 10);
    fputs("\n e = ", stderr); mp_writefile(b, stderr, 10);
    fputs("\n r = ", stderr); mp_writefile(r, stderr, 10);
    fputs("\nmr = ", stderr); mp_writefile(mr, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  MP_DROP(m);
  MP_DROP(a);
  MP_DROP(b);
  MP_DROP(r);
  MP_DROP(mr);
  mpmont_destroy(&mm);
  return ok;
}


static test_chunk tests[] = {
  { "create", tcreate, { &type_mp, &type_mp, &type_mp, &type_mp } },
  { "mul", tmul, { &type_mp, &type_mp, &type_mp, &type_mp } },
  { "exp", texp, { &type_mp, &type_mp, &type_mp, &type_mp } },
  { 0, 0, { 0 } },
};

int main(int argc, char *argv[])
{
  sub_init();
  test_run(argc, argv, tests, SRCDIR "/tests/mpmont");
  return (0);
}

#endif

/*----- That's all, folks -------------------------------------------------*/
